import os
import pathlib
import subprocess


def tm_align(pdb_file_1, pdb_file_2,out_dir):
    import os
    import subprocess
    url = 'https://zhanggroup.org/TM-score/TMscore.cpp'

    if (os.path.exists(r'TMalign')) == False:
        import requests
        r = requests.get(url, allow_redirects=True)
        open('TMalign.cpp', 'wb').write(r.content)
        print(subprocess.getoutput('g++ -O3 -ffast-math -lm -o TMalign {s}'.format(s='TMalign.cpp')))
    out_name = '{}TO{}'.format(pdb_file_1[:4], pdb_file_2[:4])
    batcmd = './TMalign {file1} {file2} -o {out_dir}/TM.{out_name}.sup -m {out_dir}/matrix{out_name}.txt'.format(file1=pdb_file_1,
                                                                                             file2=pdb_file_2,
                                                                                             out_name=out_name,
                                                                                             out_dir=out_dir)
    result = subprocess.getoutput(batcmd)
    return result



def tm_score( pdb_file_1,pdb_file_2,out_dir):
    import os
    import subprocess
    url = 'https://zhanggroup.org/TM-score/TMscore.cpp'

    if (os.path.exists(r'TMalign')) == False:
        import requests
        r = requests.get(url, allow_redirects=True)
        open('TMscore.cpp', 'wb').write(r.content)
        print(subprocess.getoutput('g++ -O3 -ffast-math -lm -o TMscore {s}'.format(s='TMscore.cpp')))
    # out_name = '{}TO{}'.format(pdb_file_1[:4], pdb_file_2[:4])
    batcmd = './TMscore {file1} {file2} '.format(file1=pdb_file_1,
                                                                                             file2=pdb_file_2,
                                                                                         )
    result = subprocess.getoutput(batcmd)
    return result


import numpy as np
import matplotlib.pyplot as plt


class BubbleChart:
    def __init__(self, area, bubble_spacing=0):
        """
        Setup for bubble collapse.

        Parameters
        ----------
        area : array-like
            Area of the bubbles.
        bubble_spacing : float, default: 0
            Minimal spacing between bubbles after collapsing.

        Notes
        -----
        If "area" is sorted, the results might look weird.
        """
        area = np.asarray(area)
        r = np.sqrt(area / np.pi)

        self.bubble_spacing = bubble_spacing
        self.bubbles = np.ones((len(area), 4))
        self.bubbles[:, 2] = r
        self.bubbles[:, 3] = area
        self.maxstep = 2 * self.bubbles[:, 2].max() + self.bubble_spacing
        self.step_dist = self.maxstep / 2

        # calculate initial grid layout for bubbles
        length = np.ceil(np.sqrt(len(self.bubbles)))
        grid = np.arange(length) * self.maxstep
        gx, gy = np.meshgrid(grid, grid)
        self.bubbles[:, 0] = gx.flatten()[:len(self.bubbles)]
        self.bubbles[:, 1] = gy.flatten()[:len(self.bubbles)]

        self.com = self.center_of_mass()

    def center_of_mass(self):
        return np.average(
            self.bubbles[:, :2], axis=0, weights=self.bubbles[:, 3]
        )

    def center_distance(self, bubble, bubbles):
        return np.hypot(bubble[0] - bubbles[:, 0],
                        bubble[1] - bubbles[:, 1])

    def outline_distance(self, bubble, bubbles):
        center_distance = self.center_distance(bubble, bubbles)
        return center_distance - bubble[2] - \
            bubbles[:, 2] - self.bubble_spacing

    def check_collisions(self, bubble, bubbles):
        distance = self.outline_distance(bubble, bubbles)
        return len(distance[distance < 0])

    def collides_with(self, bubble, bubbles):
        distance = self.outline_distance(bubble, bubbles)
        idx_min = np.argmin(distance)
        return idx_min if type(idx_min) == np.ndarray else [idx_min]

    def collapse(self, n_iterations=50):
        """
        Move bubbles to the center of mass.

        Parameters
        ----------
        n_iterations : int, default: 50
            Number of moves to perform.
        """
        for _i in range(n_iterations):
            moves = 0
            for i in range(len(self.bubbles)):
                rest_bub = np.delete(self.bubbles, i, 0)
                # try to move directly towards the center of mass
                # direction vector from bubble to the center of mass
                dir_vec = self.com - self.bubbles[i, :2]

                # shorten direction vector to have length of 1
                dir_vec = dir_vec / np.sqrt(dir_vec.dot(dir_vec))

                # calculate new bubble position
                new_point = self.bubbles[i, :2] + dir_vec * self.step_dist
                new_bubble = np.append(new_point, self.bubbles[i, 2:4])

                # check whether new bubble collides with other bubbles
                if not self.check_collisions(new_bubble, rest_bub):
                    self.bubbles[i, :] = new_bubble
                    self.com = self.center_of_mass()
                    moves += 1
                else:
                    # try to move around a bubble that you collide with
                    # find colliding bubble
                    for colliding in self.collides_with(new_bubble, rest_bub):
                        # calculate direction vector
                        dir_vec = rest_bub[colliding, :2] - self.bubbles[i, :2]
                        dir_vec = dir_vec / np.sqrt(dir_vec.dot(dir_vec))
                        # calculate orthogonal vector
                        orth = np.array([dir_vec[1], -dir_vec[0]])
                        # test which direction to go
                        new_point1 = (self.bubbles[i, :2] + orth *
                                      self.step_dist)
                        new_point2 = (self.bubbles[i, :2] - orth *
                                      self.step_dist)
                        dist1 = self.center_distance(
                            self.com, np.array([new_point1]))
                        dist2 = self.center_distance(
                            self.com, np.array([new_point2]))
                        new_point = new_point1 if dist1 < dist2 else new_point2
                        new_bubble = np.append(new_point, self.bubbles[i, 2:4])
                        if not self.check_collisions(new_bubble, rest_bub):
                            self.bubbles[i, :] = new_bubble
                            self.com = self.center_of_mass()

            if moves / len(self.bubbles) < 0.1:
                self.step_dist = self.step_dist / 2

    def plot(self, ax, labels, colors):
        """
        Draw the bubble plot.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
        labels : list
            Labels of the bubbles.
        colors : list
            Colors of the bubbles.
        """
        for i in range(len(self.bubbles)):
            circ = plt.Circle(
                self.bubbles[i, :2], self.bubbles[i, 2], color=colors[i])
            ax.add_patch(circ)
            ax.text(*self.bubbles[i, :2], labels[i],
                    horizontalalignment='center', verticalalignment='center')


import os

from typing import Optional

from tqdm import tqdm
import torch
from transformers.pipelines.base import Dataset

from ._base import BaseMLTask
from transformers import AutoTokenizer, BertForSequenceClassification

from ._io import GoAnnotationOutput
from transformers import FeatureExtractionPipeline

SUPPORTED_BERT_MODELS = ['prot_bert']
MODELS_MAP = {
    'prot_bert': 'Rostlab/prot_bert'
}


class GoADataset(Dataset):
    def __init__(self, sequences):
        self.seqs = sequences

    def __getitem__(self, index):
        return self.seqs[index]

    def __len__(self):
        return len(self.seqs)


class GoAnnotation(BaseMLTask):
    def __init__(self, model, tokenizer, num_labels, use_gpu):
        super().__init__(model, tokenizer)
        self.num_labels = num_labels
        self.use_gpu = use_gpu and torch.cuda.is_available()
        self.pipeline = None

    @classmethod
    def from_pretrained(cls, model: str, version: str, num_labels: int, use_gpu: Optional[bool] = True):
        model_name = model

        if model_name in SUPPORTED_BERT_MODELS:
            if os.path.exists(version):
                model = BertForSequenceClassification.from_pretrained(os.path.abspath(version), num_labels=num_labels)
            else:
                model = BertForSequenceClassification.from_pretrained(model_name, num_labels=num_labels)
            tokenizer = AutoTokenizer.from_pretrained(MODELS_MAP[model_name])
        else:
            raise Exception('Invalid model or version passed!')

        return cls(model, tokenizer, num_labels, use_gpu=use_gpu)

    def fit(self, *args, **kwargs):
        raise NotImplementedError

    def run(self, dataset, labels: Optional[list] = None,
            return_last_hidden_state: Optional[bool] = True, **kwargs):

        sequences = list(dataset['Sequence'].values)
        sequences = [' '.join(seq) for seq in sequences]

        goa_dataset = GoADataset(sequences)

        # def database(seqs):
        #     encoded_sequence = self.tokenizer(seqs, padding=True, return_tensors='pt')
        #     encoded_sequence = {key: value.to(self.model.device) for key, value in encoded_sequence.items()}
        #     return encoded_sequence
        #
        # model_outputs = []
        # batches = [sequences[i: i + batch_size] for i in range(0, len(sequences), batch_size)]
        # for batch in tqdm(batches):
        #     encoded = database(batch)
        #     with torch.no_grad():
        #         model_outputs.append(self.model(**encoded, output_hidden_states=return_last_hidden_state))
        if self.pipeline is None:
            self.pipeline = FeatureExtractionPipeline(model=self.model.bert, tokenizer=self.tokenizer,
                                                      task='GoA', device=0 if self.use_gpu else -1)
        for output in tqdm(self.pipeline(goa_dataset, batch_size=kwargs.get('batch_size', 32), total=len(goa_dataset))):
            pass

        return GoAnnotationOutput(outputs=output, obs=dataset.copy())



import os.path
from typing import Optional

import torch.cuda

from ._base import BaseMLTask
from transformers import AutoModelForCausalLM, AutoTokenizer, TextGenerationPipeline

from ._io import CausalLMOutput

SUPPORTED_BERT_MODELS = ['ProtGPT2']
MODELS_MAP = {
    'ProtGPT2': 'nferruz/ProtGPT2'
}


class AutoRegressiveLanguageModeling(BaseMLTask):
    def __init__(self, model, tokenizer, use_gpu: Optional[bool] = False):
        super().__init__(model, tokenizer)
        self.use_gpu = use_gpu and torch.cuda.is_available()
        self.pipeline = None

    @classmethod
    def from_pretrained(cls, model: str, version: str, use_gpu: Optional[bool] = False):
        model_name = model

        if model_name in SUPPORTED_BERT_MODELS:
            if os.path.exists(version):
                model = AutoModelForCausalLM.from_pretrained(os.path.abspath(version))
            else:
                model = AutoModelForCausalLM.from_pretrained(MODELS_MAP[model_name])
            tokenizer = AutoTokenizer.from_pretrained(MODELS_MAP[model_name])
        else:
            raise Exception('Invalid model or version passed!')

        return cls(model, tokenizer, use_gpu=use_gpu)

    def fit(self, *args, **kwargs):
        pass

    def run(self, num_samples: int = 32, batch_size: int = 32, **kwargs):
        if self.pipeline is None:
            self.pipeline = TextGenerationPipeline(model=self.model, tokenizer=self.tokenizer, batch_size=batch_size,
                                                   device=0 if self.use_gpu else -1, task='LM')
        outputs = self.pipeline('', max_length=kwargs.pop('max_length', 512), do_sample=True,
                                top_k=kwargs.pop('top_k', 950),
                                repetition_penalty=kwargs.pop('repitition_penalty', 1.2),
                                num_return_sequences=num_samples,
                                eos_token_id=kwargs.pop('eos_token_id', 0), **kwargs)

        return CausalLMOutput(outputs)



from ._clustering import cluster
from ._pdb import download_list_pdb, get_pdb_parser,download_from_alphafold
from ._tmalign import tm_align,tm_score

__all__ = ['cluster','download_list_pdb','get_pdb_parser','download_from_alphafold','tm_align','tm_score']



from collections import defaultdict
from typing import Optional

import scanpy as sc



class Graph:
    def __init__(self, vertices):
        self.graph = defaultdict(list)
        self.ids = defaultdict(list)
        self.V = vertices
        self.visit = {}
        self.label = {}
        self.paths = defaultdict(list)

    def add_edge(self, u, v):
        self.graph[u].append(v)
        self.ids[u] = u
        self.ids[v] = v
        self.visit[u] = False
        self.visit[v] = False

    def dfs_util(self, v, label):
        if self.label.get(v) is None:
            self.label[v] = label
        else:
            self.label[v] = max(self.label[v],label)

        for x in self.graph[v]:
            self.dfs_util(x,self.label[v]+1)



def _construct_adata(model_results, color: Optional[str] = None, **kwargs):
    adata = sc.AnnData(X=model_results.get_hidden_states(layer=kwargs.get('layer', -1)))
    adata.obs = model_results.obs.copy()

    if color:
        adata.obs['color'] = model_results.obs[color].values
    else:
        pass

    return adata


def convert_column_to_list(adata, column):
    adata.obs[column] = adata.obs[column].apply(lambda cell: ''.join(c for c in cell if c not in "\'\"[]").split(', '))


def umap(model_results, color: Optional[str] = None, plot_type: Optional[str] = None, **kwargs):
    sc.settings.set_figure_params(dpi=kwargs.pop('dpi', 100))
    if plot_type == '1_vs_all':
        adata = _construct_adata(model_results, color=None, layer=kwargs.pop('layer', -1))
    else:
        adata = _construct_adata(model_results, color, layer=kwargs.pop('layer', -1))

    sc.pp.neighbors(adata)
    sc.tl.umap(adata)

    if plot_type == '1_vs_all':
        convert_column_to_list(adata, color)

        unique_values = set()
        for group in adata.obs[color].to_list():
            unique_values |= set(group)
        unique_values = sorted(list(unique_values))

        for term in unique_values:
            adata.obs[term] = 'Others'
            adata.obs.loc[adata.obs[color].apply(lambda x: term in x), term] = term
            sc.pl.umap(adata, color=term, frameon=False, **kwargs)
    else:
        sc.pl.umap(adata, color='color' if color else None, **kwargs)




def draw_term_family(id):
    import lib310
    import graphviz
    query = """SELECT
          pathList
        FROM
          `pfsdb3.1_go.term_term_path`
        WHERE
          child_id = "{id}";""".format(id=id)
    results = lib310.db.fetch(query=query)



    results = lib310.db.fetch(query="""SELECT
      pathList
    FROM
      `pfsdb3.1_go.term_term_path`
    WHERE
      child_id = "{id}"
    """.format(id=id))
    # import graphviz
    from collections import defaultdict

    g = Graph(len(results))
    relationtype = {
        'is_a': {'color': 'black', 'style': 'bold'},
        'part_of': {'color': 'blue', 'style': 'dashed'},
        'regulates': {'color': 'darkorange', 'style': 'bold'},
        'positively_regulates': {'color': 'darkgreen', 'style': 'bold'},
        'negatively_regulates': {'color': 'red', 'style': 'bold'},
        'occurs_in': {'color': 'darkslategray4', 'style': 'bold'},
    }

    defultrel = {'color': 'gray', 'style': 'dotted'}

    nodes = {}
    u = graphviz.Digraph('unix', filename='unix.gv')
    u.attr(size='1000,1000')
    u.attr(ranksep="2")
    u.attr('node', shape='plaintext')
    u.attr('node', font='helvetica')

    res = lib310.db.fetch(query="""SELECT go_id,name FROM `pfsdb3.1_go.terms`  """)
    node_names = {}
    for row in res.values.tolist():
        node_names[row[0]] = row[1]
    edgeList = defaultdict(set)
    rel = defaultdict(dict)
    data = results[['pathList']].values.tolist()
    for row in data:
        paths = row
        pathList = paths[0].strip('][').split(', ')
        x = 0
        while (x + 2 < len(pathList)):
            edgeList[pathList[x]].add(pathList[x + 2])
            rel[pathList[x]][pathList[x + 2]] = pathList[x + 1]
            x = x + 2
    for x in edgeList:
        for y in edgeList[x]:
            z = rel[x][y].replace(' " ', " ")
            z = z.strip(' " " ')
            z = z.replace("'", '')
            X = "" + x[4:11]
            Y = "" + y[4:11]
            if X not in nodes:
                nodes[X] = node_names["GO:" + X]
            if Y not in nodes:
                nodes[Y] = node_names["GO:" + Y]
            if z in relationtype.keys():
                u.edge(x[4:11], y[4:11], color=relationtype[z]['color'], style=relationtype[z]['style'])
                g.add_edge(y[4:11], x[4:11])
            else:
                z = defultrel
                u.edge(x[4:11], y[4:11], color=z['color'], style=z['style'])
                # g.add_edge( y[4:11],x[4:11])

    g.dfs_util(id[3:], 1)

    i = 1
    while i < 1000:
        with u.subgraph() as s:
            s.attr(rank='same')

            for x in g.label.keys():
                if g.label[x] == i:
                    # s.node(x)
                    name = "GO:" + str(x)
                    s.node(x, label='''<
                    <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="2" CELLPADDING="3">
                      <TR>

                        <TD BGCOLOR="deepskyblue3" COLSPAN="1">{id}</TD>
                      </TR>

                      <TR>
                        <TD  COLSPAN="1">{name}</TD>
                      </TR>
                    </TABLE>>'''.format(id=x, name=node_names[name]))

                    # print(str(x)+': '+str(i))

        i = i + 1

    for node in nodes:
        if node not in g.label.keys():
            u.node(node, label='''<
        <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="2" CELLPADDING="3">
          <TR>

            <TD BGCOLOR="deepskyblue3" COLSPAN="1">{id}</TD>
          </TR>

          <TR>
            <TD  COLSPAN="1">{name}</TD>
          </TR>
        </TABLE>>'''.format(id=node, name=node_names["GO:" + node]))
    return u


def show_umap(features):
    from umap import UMAP
    import plotly.express as px

    umap_2d = UMAP(n_components=2, init='random', random_state=0)
    umap_3d = UMAP(n_components=3, init='random', random_state=0)

    proj_2d = umap_2d.fit_transform(features)
    proj_3d = umap_3d.fit_transform(features)

    fig_2d = px.scatter(
        proj_2d, x=0, y=1,
        color=0,
    )
    fig_3d = px.scatter_3d(
        proj_3d, x=0, y=1, z=2,
        color=0,
    )
    fig_3d.update_traces(marker_size=5)

    fig_2d.show()
    fig_3d.show()


def pca(model_results, color: Optional[str] = None, plot_type: Optional[str] = None, **kwargs):
    sc.settings.set_figure_params(dpi=kwargs.pop('dpi', 100))
    if plot_type == '1_vs_all':
        adata = _construct_adata(model_results, color=None, layer=kwargs.pop('layer', -1))
    else:
        adata = _construct_adata(model_results, color=color, layer=kwargs.pop('layer', -1))

    sc.pp.pca(adata, n_comps=50)

    if plot_type == '1_vs_all':
        convert_column_to_list(adata, color)

        unique_values = set()
        for group in adata.obs[color].to_list():
            unique_values |= set(group)
        unique_values = sorted(list(unique_values))

        for term in unique_values:
            adata.obs[term] = 'Others'
            adata.obs.loc[adata.obs[color].apply(lambda x: term in x), term] = term
            sc.pl.pca(adata, color=term, **kwargs)
    else:
        sc.pl.pca(adata, color='color' if color else None, **kwargs)


def tsne(model_results, color: Optional[str] = None, plot_type: Optional[str] = None, **kwargs):
    sc.settings.set_figure_params(dpi=kwargs.get('dpi', 100))
    if plot_type == '1_vs_all':
        adata = _construct_adata(model_results, color=None, layer=kwargs.pop('layer', -1))
    else:
        adata = _construct_adata(model_results, color, layer=kwargs.pop('layer', -1))

    sc.pp.neighbors(adata)
    sc.tl.tsne(adata)

    if plot_type == '1_vs_all':
        convert_column_to_list(adata, color)

        unique_values = set()
        for group in adata.obs[color].to_list():
            unique_values |= set(group)
        unique_values = sorted(list(unique_values))

        for term in unique_values:
            adata.obs[term] = 'Others'
            adata.obs.loc[adata.obs[color].apply(lambda x: term in x), term] = term
            sc.pl.tsne(adata, color=term, **kwargs)
    else:
        sc.pl.tsne(adata, color='color' if color else None, frameon=False, **kwargs)


def visualize_3model(fpath, chain_id):
        import py3Dmol

        with open(fpath) as ifile:
            system = "".join([x for x in ifile])

        view = py3Dmol.view(width=600, height=400)
        view.addModelsAsFrames(system)
        view.setStyle({'model': -1, 'chain': chain_id}, {"cartoon": {'color': 'spectrum'}})
        view.setBackgroundColor(0xe6ecf5)
        view.zoomTo()
        view.show()


import logging
from typing import Union

from rich.console import Console
from rich.logging import RichHandler

b10_logger = logging.getLogger("lib310")


class B10Config:
    def __init__(self, verbosity: int = logging.INFO):
        self.verbosity = verbosity

    def reset_logging_handler(self):
        """
        Resets "lib310" log handler to a basic RichHandler().
        This is useful if piping outputs to a file.
        """
        b10_logger.removeHandler(b10_logger.handlers[0])
        ch = RichHandler(level=self._verbosity, show_path=False, show_time=False)
        formatter = logging.Formatter("%(message)s")
        ch.setFormatter(formatter)
        b10_logger.addHandler(ch)

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, level: Union[int, str]):
        self._verbosity = level
        b10_logger.setLevel(level)
        if len(b10_logger.handlers) == 0:
            console = Console(force_terminal=True)
            if console.is_jupyter is True:
                console.is_jupyter = False
            ch = RichHandler(
                level=level, show_path=False, console=console, show_time=False
            )
            formatter = logging.Formatter("%(message)s")
            ch.setFormatter(formatter)
            b10_logger.addHandler(ch)
        else:
            b10_logger.setLevel(level)


settings = B10Config()



from typing import Optional

import matplotlib.pyplot as plt
import pandas as pd
from ._utils import color


class ProteinDataTable(pd.DataFrame):
    @classmethod
    def from_pandas(cls, df):
        return cls(data=df.copy())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def summary(self, verbosity: int = 1):
        if verbosity == 1:
            print(f'* Shape: ({color.BOLD}{self.shape[0]}{color.END}, {color.BOLD}{self.shape[1]}{color.END})\n')
            print(f'* Columns information:')
            for i, column in enumerate(self.columns):
                print(
                    f'\t{i + 1}. {color.BOLD}{column}{color.END}: data type {color.BLUE}{self.dtypes[column]}{color.END} with {color.BOLD}{self[column].isna().sum()}{color.END} Missing values')

            print(f'\n* Total number of missing values in the data: {color.BOLD}{self.isna().sum().sum()}{color.END}')
            print(f'* Number of rows with missing values: {color.BOLD}{self.shape[0] - self.dropna().shape[0]}{color.END}')
        else:
            raise ValueError('Wrong verbosity level has been set! You have to set it only to `1` for now!')

    def plot(self, column: str, plot_type: Optional[str] = None):
        assert column in self.columns

        import seaborn as sns
        sns.set(rc={"figure.dpi": 150, 'savefig.dpi': 150})
        sns.set_context('notebook')
        sns.set_style("ticks")

        dtype = str(self.dtypes[column]).lower()

        if dtype in ['int64', 'int32', 'int']:
            if plot_type is None or plot_type == 'hist':
                sns.displot(
                    self, x=column,
                    binwidth=100, height=3, aspect=2, facet_kws=dict(margin_titles=True), kind='hist',
                )
                plt.show()
            elif plot_type in ['dist', 'kde']:
                sns.displot(
                    self, x=column,
                    binwidth=100, height=3, aspect=2, facet_kws=dict(margin_titles=True), kind='kde',
                )
                plt.show()
        elif dtype in ['object', 'str'] and plot_type == 'population':
            self['temp'] = self[column].apply(lambda cell: ''.join(c for c in cell if c not in "\'\"[]").split(', '))

            all_terms = set()
            for term_group in self['temp'].to_list():
                all_terms |= set(term_group)
            all_terms = sorted(list(all_terms))

            print(f'Found {len(all_terms)} different values')

            term_population = {column: [], 'population': []}
            for term in all_terms:
                term_df = self.loc[self['temp'].apply(lambda x: term in x)]

                term_population[column].append(term)
                term_population['population'].append(len(term_df))

            plt.figure(figsize=(15, 6))
            sns.barplot(x=column, y='population', palette="deep", data=pd.DataFrame.from_dict(term_population))
            plt.xticks(rotation=90)
            plt.grid(axis='y')
            plt.show()
        else:
            print('Invalid function usage')



from google.cloud.bigquery import DestinationFormat
class CONSTATNS:
    SEQUENCE_KEY = 'Sequence'


class FileFormat(object):
    CSV = DestinationFormat.CSV
    JSON = DestinationFormat.NEWLINE_DELIMITED_JSON
    AVRO = DestinationFormat.AVRO
    PARQUET = DestinationFormat.PARQUET

    @staticmethod
    def to_extension(value):
        extensions = {
            FileFormat.CSV: 'csv',
            FileFormat.JSON: 'json',
            FileFormat.AVRO: 'avro',
            FileFormat.PARQUET: 'parquet'
        }
        return extensions[value]

    @staticmethod
    def to_format(value):
        value = value.upper()
        formats = {
            'CSV': FileFormat.CSV,
            'JSON': FileFormat.JSON,
            'AVRO': FileFormat.AVRO,
            'PARQUET': FileFormat.PARQUET,
            'JSONNL': FileFormat.JSON,
            FileFormat.JSON: FileFormat.JSON
        }
        return formats[value]


class CacheResponseType(object):
    CACHE_INFO = 'cache_info'
    DATASET = 'dataset'


from abc import ABC, abstractmethod


class BaseMLTask(ABC):
    def __init__(self, model, tokenizer, **kwargs):
        self.model = model
        self.tokenizer = tokenizer

    @abstractmethod
    def fit(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def run(self, *args, **kwargs):
        raise NotImplementedError



import random
import string

def number_to_abbrevation(num):
    label = ['', 'K', 'M', 'B', 'T']
    i = 0
    while num > 1000:
        num /= 1000
        i += 1
    return f'{num:.1f} {label[i]}'


def size_to_abbrevation(num):
    label = ['B', 'KB', 'MB', 'GB', 'TB']
    i = 0
    while num > 1024:
        num /= 1024
        i += 1
    return f'{num:.1f} {label[i]}'


def get_random_string(length):
    return ''.join(random.choice(string.ascii_uppercase) for i in range(length))


from pymongo import MongoClient, ASCENDING
import random
import time
import pandas as pd
from threading import Thread, Event
from queue import Queue


class MLDL:
    def __init__(self, mongo_url, cache_size=10):
        random.seed(time.time())
        self.client = MongoClient(mongo_url)
        self.db = self.client['A310']
        self.col_train = self.db['seq_lang5_train']
        self.col_test = self.db['seq_lang5_test']
        self.cache = {'TRAIN': {}, 'TEST': {}}
        self.sample_cache = []
        self.max_queue_size = cache_size
        self.queue_key = None
        self.queue = Queue(self.max_queue_size)
        self.filler_thread = None
        self.kill_event = Event()
        self.kill_event_cache = Event()
        self.filler_thread_cache = None

        tmp = self.db['length_freq_lang5_train'].find().sort('len', ASCENDING)
        for item in tmp:
            self.cache['TRAIN'][item['len']] = {'freq': item['freq'], 'maxi': item['maxi'], 'mini': item['mini']}
        tmp = self.db['length_freq_lang5_test'].find().sort('len', ASCENDING)
        for item in tmp:
            self.cache['TEST'][item['len']] = {'freq': item['freq'], 'maxi': item['maxi'], 'mini': item['mini']}

    def get_batch(self, num, max_length, min_num_feature=0, stage='TRAIN'):
        stage = stage.upper()
        if stage == 'TRAIN':
            col = self.col_train
        elif stage == 'TEST':
            col = self.col_test
        else:
            raise ValueError('Stage must be either TRAIN or TEST')
        if max_length not in self.cache[stage.upper()].keys():
            arr = [i for i in self.cache[stage.upper()].keys() if i <= max_length]
            if (len(arr) == 0):
                return pd.DataFrame()
            max_length = max(arr)

        hit = self.queue_key == f'{num}_{max_length}_{min_num_feature}_{stage}'
        if hit:
            # HIT
            return pd.DataFrame(self.queue.get())

        # MISS
        maxi = self.cache[stage.upper()][max_length]['maxi']

        if self.filler_thread_cache is not None:
            self.kill_event_cache.set()
            self.filler_thread_cache.join()
            self.kill_event_cache.clear()
            self.filler_thread_cache = None

        if min_num_feature > 11:
            c = col.find({'len': {'$lt': max_length + 1}, 'token_ids': {'$size': {'$gt': min_num_feature - 1}}},
                         {'row_id': 1, '_id': 0}).sort("rand").limit(5 * num)
            self.sample_cache = list(map(lambda x: x['row_id'], list(c)))
            self.filler_thread_cache = Thread(target=self.background_sample_cache,
                                              args=(max_length, num, min_num_feature, col))
            self.filler_thread_cache.start()
        else:
            self.sample_cache = range(1, maxi)

        if self.filler_thread is not None:
            self.kill_event.set()
            self.queue.get()
            self.filler_thread.join()
            self.kill_event.clear()
            self.queue = Queue(self.max_queue_size)

        self.queue_key = f'{num}_{max_length}_{min_num_feature}_{stage}'

        self.filler_thread = Thread(target=self.filler, args=(num, col))
        self.filler_thread.start()

        index_list = random.sample(self.sample_cache, min(num, len(self.sample_cache)))
        cursor = col.find({'row_id': {'$in': index_list}}, {'_id': 1, 'sequence': 1, 'token_ids': 1})
        return pd.DataFrame(list(cursor))

    def filler(self, num, col):
        while True:
            index_list = random.sample(self.sample_cache, min(num, len(self.sample_cache)))
            cursor = col.find({'row_id': {'$in': index_list}}, {'_id': 1, 'sequence': 1, 'token_ids': 1})
            df = pd.DataFrame(list(cursor))
            self.queue.put(df)
            if self.kill_event.is_set():
                break

    def background_sample_cache(self, max_length, num, min_num_feature, col):
        i = 5
        while i > 0:
            c = col.find({'len': {'$lt': max_length + 1}, 'token_ids': {'$size': {'$gt': min_num_feature - 1}}},
                         {'row_id': 1, '_id': 0}).sort("rand").skip(i * num).limit(num)
            tmp = list(c)
            self.sample_cache += list(map(lambda x: x['row_id'], tmp))
            i += 1
            if len(tmp) < num or self.kill_event_cache.is_set():
                i = 0
                break

    def terminate(self):
        self.kill_event.set()
        self.queue.get()
        self.filler_thread.join()
        self.kill_event.clear()
        self.kill_event_cache.set()
        self.filler_thread_cache.join()
        self.kill_event_cache.clear()
        self.filler_thread_cache = None


import requests

BASE_URL = "https://files.rcsb.org/download"


def download(pdb_id, file_type, out_dir):
    if '_' in pdb_id:
        pdb_id = pdb_id[:-2]
    url = "{BASE_URL}/{pdb_id}.{type}.gz".format(pdb_id=pdb_id, BASE_URL=BASE_URL, type=file_type)
    res = requests.get(url)
    outfile = "{out_dir}/{pdb_id}.{type}.gz".format(pdb_id=pdb_id, type=file_type, out_dir=out_dir)
    open(outfile, "wb").write(res.content)
    import gzip
    import shutil
    zipfile = outfile
    output_file = zipfile[:-3]
    with gzip.open(zipfile, 'rb') as f_in:
        with open(output_file, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)

    import os
    os.remove(outfile)


def download_list_pdb(pdb_list, type_list, out_dir):
    import os
    os.makedirs(out_dir, exist_ok=True)
    for pdb in pdb_list:
        for tp in type_list:
            download(pdb, tp, out_dir)


def get_pdb_parser(pdb_name, pdb_file):
    from Bio.PDB import PDBParser
    return PDBParser().get_structure(pdb_name, pdb_file)


# download_list_pdb(["7niu_A", "6s7p_A"], ["pdb", "pdb1", "xml"],"./data")


def download_from_alphafold(source_blob_name, destination_file_name):
    from google.cloud import storage
    query = """with file_rows AS (
      with file_cols AS (
        SELECT
          CONCAT(entryID, '-model_v', latestVersion, '.cif') as m,
          CONCAT(entryID, '-predicted_aligned_error_v', latestVersion, '.json') as p
        FROM bigquery-public-data.deepmind_alphafold.metadata
        WHERE organismScientificName = "Homo sapiens"
          AND (fractionPlddtVeryHigh + fractionPlddtConfident) > 0.5
      )
      SELECT * FROM file_cols UNPIVOT (files for filetype in (m, p))
    )
    SELECT CONCAT('gs://public-datasets-deepmind-alphafold/', files) as files
    FROM file_rows LIMIT 10"""
    bucket_name = "public-datasets-deepmind-alphafold"
    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(source_blob_name)
    blob.download_to_filename(destination_file_name)
    print(
        "Downloaded storage object {} from bucket {} to local file {}.".format(
            source_blob_name, bucket_name, destination_file_name
        )
    )
# download_blob("public-datasets-deepmind-alphafold","AF-J3QLM0-F1-model_v3.cif","a.cif")



import matplotlib.pyplot as plt
import math
from .charts.bubble_chart import BubbleChart
import plotly.express as px
import pandas as pd
from ._functions import *



def visualize_all_datasets(df):
    import seaborn as sns # added here for lazy load
    bubble_chart = BubbleChart(area=[math.log(x) for x in df['num_rows']],
                               bubble_spacing=0.1)
    bubble_chart.collapse()

    fig, ax = plt.subplots(subplot_kw=dict(aspect="equal"))
    bubble_chart.plot(ax, list(df.index), sns.color_palette("Spectral", len(list(df.index))).as_hex())
    ax.axis("off")
    ax.relim()
    ax.autoscale_view()
    ax.set_title('Dataset (All)')

    plt.show()


def visualize_dataset(df, name):
    import seaborn as sns # added here for lazy load
    fig, ax = plt.subplots()
    ax.bar(df['table_name'], df['num_rows'], color=sns.color_palette("Spectral", len(df['table_name'])).as_hex())
    ax.axis("on")
    ax.relim()
    ax.autoscale_view()
    ax.set_title('Dataset (' + name + ')')
    plt.yscale('log')
    plt.xlabel('Table')
    plt.ylabel('number of rows')
    plt.show()


def visualize_all(df):
    df = pd.DataFrame({'310 db': ['310 DB'] * len(df['dataset']),
                       'dataset': df['dataset_name'],
                       'table': df['table_name'],
                       'bg_color': [item['color'] for item in df['treemap']],
                       'textcolor': [item['textcolor'] for item in df['treemap']],
                       'size': [size_to_abbrevation(n) for n in df['size']],
                       'weight': df['num_rows'].apply(lambda x: math.log10(x) - 4 if math.log10(x) - 4 > 1 else 1),
                       'rows': [number_to_abbrevation(n) for n in df['num_rows']]
                       })
    leaves = df.select_dtypes("object").apply("/".join, axis=1).values

    fig = px.treemap(df, path=['310 db', 'dataset', 'table'], values='weight', hover_data=["size", "rows"])
    bg_colors = []
    text_colors = []
    for sector in fig.data[0]['ids']:
        if len(sector.split('/')) == 3:
            idx = df.index[df['dataset'] == sector.split('/')[1]]
            bg_colors.append(df.loc[idx, 'bg_color'].values[0])
            text_colors.append(df.loc[idx, 'textcolor'].values[0])
            continue
        bg_colors.append('transparent')

    fig.data[0]['marker']['colors'] = bg_colors
    fig.data[0].textposition = 'middle center'
    fig.data[0]['marker']['line']['color'] = '#D3D3D3'
    fig.data[0].texttemplate = "<span style='font-size: 16px;'> %{label} </span> <br> <span style='font-size: 12px;'>%{customdata[0]} <br> %{customdata[1]}<span>"
    fig.data[0]['textfont']['color'] = text_colors
    fig.update_layout(margin=dict(t=50, l=25, r=25, b=25))
    # fig.update_traces(hovertemplate='%{customdata[0]}')
    # fig.data[0].customdata = [
    #     v if fig.data[0].ids[i]
    #          in leaves else [""]
    #     for i, v in enumerate(fig.data[0].customdata)
    # ]
    fig.show()


from google.cloud import storage
import google.auth
import gcsfs
from torch.utils.data import Dataset
import dask.dataframe as dd
from ._constants import FileFormat
import threading



class GCSDataset(Dataset):
    def __init__(self, file_uri, target_col_name, file_format=FileFormat.CSV, size=-1):
        """
        :param file_uri: the uri of file even with wildcards in gcs
        :return:
        """
        if not size:
            size = -1
        storage_client = storage.Client()
        credentials, project = google.auth.default(scopes=storage_client.SCOPE)
        fs = gcsfs.GCSFileSystem(project=project, token=credentials)
        fs.credentials.maybe_refresh()
        file_format = FileFormat.to_format(file_format)
        if file_format == FileFormat.CSV:
            self.ddf = dd.read_csv(file_uri, storage_options={'token': fs.credentials.token})
        elif file_format == FileFormat.JSON:
            self.ddf = dd.read_json(file_uri, storage_options={'token': fs.credentials.token})
        elif file_format == FileFormat.PARQUET:
            self.ddf = dd.read_parquet(file_uri, storage_options={'token': fs.credentials.token})
        else:
            raise TypeError('The file_format is not supported. supported("csv", "json", "parquet")')
        self.target_name = target_col_name
        self.file_uri = file_uri
        self.lock = threading.Lock()


        if size < 0:
            self.n = 0
            self.parts = []
            for i in range(self.ddf.npartitions):
                df = self.ddf.partitions[i].compute()
                self.n = self.n + len(df)
                self.parts.append({
                    'fetched': True,
                    'start': self.n - len(df),
                    'end': self.n - 1
                })
            self.last_partition_loaded = self.ddf.get_partition(0).compute()
            self.last_partition_loaded_index = 0
        else:
            self.last_partition_loaded = self.ddf.get_partition(0).compute()
            self.last_partition_loaded_index = 0
            first_partition_size = len(self.last_partition_loaded)
            self.n = size
            self.parts = [
                {
                    'fetched': (i == 0),
                    'start': (i * first_partition_size),
                    'end': min(size, ((i + 1) * first_partition_size) - 1)
                } for i in range(self.ddf.npartitions)]

    def __len__(self):
        return self.n

    def __getitem__(self, idx):
        try:
            _, part_idx, df = self.find_correct_dask_partition(idx)
            if part_idx >= len(df):
                print('idx: ', idx)
                return None
            item = df.iloc[part_idx, df.columns.get_indexer(df.columns[df.columns != self.target_name])]
            if self.target_name is not None and self.target_name != '' and self.target_name in df.columns:
                target = df.iloc[part_idx, df.columns.get_indexer(df.columns[df.columns == self.target_name])]
                target = target.to_list()
                return dict(item), dict(target)
            return dict(item)
        except Exception as e:
            print(e)
            return None

    def find_correct_dask_partition(self, idx):
        self.lock.acquire()
        for i, part in enumerate(self.parts):
            if not part['fetched']:
                prv = self.parts[i - 1]
                if self.last_partition_loaded_index != i:
                    self.last_partition_loaded = self.ddf.partitions[i].compute()
                    self.last_partition_loaded_index = i
                tmp = {
                    'fetched': True,
                    'start': prv['end'] + 1,
                    'end': prv['end'] + len(self.last_partition_loaded)
                }
                self.parts[i] = tmp

            if part['start'] <= idx <= part['end']:
                if self.last_partition_loaded_index != i:
                    lp = self.ddf.partitions[i].compute()
                    li = i
                    self.last_partition_loaded = lp
                    self.last_partition_loaded_index = li
                else:
                    li = self.last_partition_loaded_index
                    lp = self.last_partition_loaded
                part_idx = idx - part['start']
                result = (li, part_idx, lp)
                self.lock.release()
                return result

        part_idx = idx - self.parts[-1]['start']
        li = self.last_partition_loaded_index
        lp = self.last_partition_loaded
        result = (li, part_idx, lp)
        self.lock.release()
        return result



import os


def set_gcloud_key_path(path: str):
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = path


class DatabaseConnection(object):
    def __init__(self):
        from .client import Client  # This import is here for lazy load

        if "GOOGLE_APPLICATION_CREDENTIALS" not in os.environ.keys():
            raise Exception(
                'You must set BigQuery credentials first! try using `lib310.db.set_gcloud_key_path(path) method.`')

        self.client = Client()
        # self.table_name = kwargs.get('table_name', 'pfsdb3.0_uniprot.uniref')
        # self.table = self.client.get_table(self.table_name) if self.table_name else None
        # self.table_schema = {schema.name: {'type': schema.field_type, } for schema in
        #                      self.table.schema} if self.table else {}
        # self.verbose = kwargs.pop('verbose', False)
        #
        # if self.verbose:
        #     print(f'Successfully established a connection with lib310 Database!')

    def get_table_info(self):
        for column_name, column_info in self.table_schema.items():
            print(f'{column_name}:{column_info["type"]}')

    def query(self, *args, **query_params):
        from google.cloud import bigquery # This import is here for lazy load
        from ..data._base import ProteinDataTable # This import is here for lazy load
        """
        :param args: can have a string which is query
        :param query_params: if query not mentioned we read other query parameters other wise we only read query
        :param - query
        :param - table_name :default=pfsdb3.0_go.gaf
        :param - limit      :default=100
        :return:
        """

        has_query = False
        if 'query' in query_params.keys():
            has_query = True
            query_results = self.client.query(query_params.get('query')).result().to_dataframe()
        if not has_query and len(args) > 0:
            has_query = True
            query_results = self.client.query(args[0]).result().to_dataframe()

        if not has_query:
            table_name = query_params.pop('table_name', 'pfsdb3.0_go.gaf')
            if table_name != self.table_name:
                self.table_name = table_name
                self.table = self.client.get_table(table_name)
                self.table_schema = {schema.name: {'type': schema.field_type, } for schema in
                                     self.table.schema} if self.table else {}
                if self.verbose:
                    print(f'Successfully connected to {table_name} with {self.table.num_rows} rows!')

            query_string = f"SELECT * FROM `{table_name}`"

            force_limit = query_params.pop('force_limit', True)
            limit = query_params.pop('limit', 100)

            query_parameters = []
            for key, query_value in query_params.items():
                if key in self.table_schema.keys():
                    if not isinstance(query_value, list):
                        query_value = [query_value]
                    query_parameters.append(
                        bigquery.ArrayQueryParameter(f'{key}s', self.table_schema[key]['type'], query_value))

            job_config = bigquery.QueryJobConfig(
                query_parameters=query_parameters,
            )

            constraints = 0
            for key, query_value in query_params.items():
                if key in self.table_schema.keys():
                    if constraints == 0:
                        query_string += ' WHERE '

                    query_string += f'{key} IN UNNEST(@{key}s)'

                    if constraints >= 0:
                        query_string += ' AND '
                    constraints += 1

            query_string = query_string[:-5]

            if force_limit:
                query_string += f' LIMIT {limit}'

            print(query_string)
            query_results = self.client.query(query_string, job_config=job_config).result().to_dataframe()

        query_results = ProteinDataTable.from_pandas(query_results)
        return query_results



from ._connection import DatabaseConnection, set_gcloud_key_path
from ._visualize import *
from ._constants import FileFormat, CacheResponseType
from datetime import datetime, timedelta
from .gcs_dataset import GCSDataset
import logging as log
import hashlib
from .mldl import MLDL


db_connection: DatabaseConnection = None
__db_info = 'system.info'
__db_gcs_cache = 'system.gcs_cache'

__all__ = ['fetch', 'get_table_info', 'list_datasets', 'list_tables', 'summary', 'visualize', 'CacheResponseType', 'MLDL']
def fetch(*args, **kwargs):
    global db_connection
    if db_connection is None:
        db_connection = DatabaseConnection()

    query_results = db_connection.query(*args, **kwargs)
    return query_results


def get_table_info(**kwargs):
    global db_connection
    if db_connection is None:
        db_connection = DatabaseConnection()
    elif db_connection.table_name != kwargs.get('table_name', db_connection.table_name):
        db_connection = DatabaseConnection()

    return db_connection.get_table_info()


def list_datasets(**kwargs):
    global db_connection
    if db_connection is None:
        db_connection = DatabaseConnection()
    return list(db_connection.client.list_datasets())


def list_tables(**kwargs):
    global db_connection
    if db_connection is None:
        db_connection = DatabaseConnection()

    dataset_ids = kwargs.get('dataset_ids')

    if dataset_ids is None:
        dataset_ids = [d.dataset_id for d in list(db_connection.client.list_datasets())]

    if not isinstance(dataset_ids, list):
        dataset_ids = [dataset_ids]
        return list(db_connection.client.list_tables(dataset_ids[0]))

    result = {}
    for dataset_id in dataset_ids:
        result[dataset_id] = list(db_connection.client.list_tables(dataset_id))
    return result


def summary(**kwargs):
    global db_connection
    db_connector()
    client = db_connection.client
    df = client.query(f"SELECT dataset_name, table_name, num_rows, num_cols, size FROM `{__db_info}`").to_dataframe()

    print_it = kwargs.get('print')
    if print_it is None or print_it is True:
        df['size'] = df['size'].apply(size_to_abbrevation)
        print(df.to_string(index = False))
    return df


def visualize(**kwargs):
    name = kwargs.get('dataset')
    global db_connection
    db_connector()
    client = db_connection.client
    if name is None:
        df = client.query(f"""
                SELECT dataset, dataset_name, table, table_name, num_rows, num_cols, size, treemap
                FROM `{__db_info}`
                WHERE treemap.show = true AND num_rows > 0""").to_dataframe()

        visualize_all(df)
        return
    df = client.query(f"SELECT dataset_name, table_name, num_rows, num_cols, size FROM `{__db_info}`").to_dataframe()
    tables = df.where(df['dataset_name'] == name.upper()).dropna()
    if len(tables) <= 0:
        return
    visualize_dataset(tables, name)


def db_connector():
    global db_connection
    if db_connection is None:
        db_connection = DatabaseConnection()
    return db_connection


def cache_query(query,
                name: str = None,
                destination_format: str or FileFormat = FileFormat.CSV,
                days: int = 7,
                ignore_hit: bool = False,
                response_type: CacheResponseType = CacheResponseType.CACHE_INFO, target_column=''):
    """
    Cache a query result in Google Cloud Storage (GCS)
    :param query: query to run on bigquery and cache in GCS
    :param name: name of the file to store in GCS
    :param destination_format: format of the file to store in GCS
    :param days: days to keep the file in GCS
    :param ignore_hit: ignore cache hit and run the query again
    :param response_type: return response as cache info or as a dataset
    :param target_column: target column to use for the dataset
    :return: cache info or dataset
    """

    if response_type == CacheResponseType.DATASET and FileFormat.to_format(destination_format) == FileFormat.AVRO:
        raise ValueError("Cannot return dataset for AVRO format")
    if str is None:
        name = get_random_string(10)

    hashed_query = hashlib.sha1(query.encode('utf-8')).hexdigest()
    if days <= 0:
        days = 7
    db = db_connector()

    # check for the hit
    row = None
    if not ignore_hit:
        cached = db.client.query('SELECT * from `system.gcs_cache` WHERE status_code != -1').to_dataframe()
        if len(cached) > 0:
            cached = cached.where(cached['hash'] == hashed_query).dropna().sort_values(by=['created_at'],
                                                                                       ascending=False)
            if len(cached) > 0:
                row = cached.iloc[0].to_dict()
    if row is not None:
        # hit happened
        row['hit'] = True
        if response_type == CacheResponseType.CACHE_INFO:
            return row
        return GCSDataset(row['uri'], target_col_name=target_column,
                          file_format=FileFormat.to_format(destination_format), size=row['total_rows'])


    table = db.client.build_temp_table()
    log.debug(f'Created table {table.table_id} in {db.client.CACHE_DATASET}')

    qresult = db.client.query_to_cached_dataset(query=query, destination=table)
    log.debug(f'Insert query {query} to {table.table_id}')

    destination_format = destination_format.upper()
    try:
        destination_format = FileFormat.to_format(destination_format)
    except KeyError:
        raise TypeError('The destination format is not supported')

    res = db.client.export_to_gcs(table, name, destination_format)
    log.debug(f'Export {table.table_id} to GCS(/{table.table_id}/{name}_*.{destination_format})')

    try:
        row = {
            'name': name,
            'folder': table.table_id,
            'uri': res.destination_uris[0],
            'length': res.destination_uri_file_counts[0],
            'created_at': datetime.now().isoformat(),
            'expired_at': (datetime.now() + timedelta(days=days)).isoformat(),
            'query': query,
            'hash': hashed_query,
            'total_rows': qresult.total_rows,
            'status_code': 1
        }
        info = db.client.insert_rows_json(__db_gcs_cache, [row])
        if len(info) > 0 and len(info[0]['errors']) != 0:
            log.error(info[0]['errors'])
    except Exception as e:
        log.error(e)

    db.client.delete_table(table)
    log.debug(f'Deleted table {table.table_id} from {db.client.CACHE_DATASET}')

    row['hit'] = False

    if response_type == CacheResponseType.CACHE_INFO:
        return row
    return GCSDataset(row['uri'], target_col_name=target_column, file_format=FileFormat.to_format(destination_format), size=row['total_rows'])




from typing import Optional

import numpy as np

from lib310.data import ProteinDataTable


class GoAnnotationOutput(object):
    def __init__(self, outputs, obs: ProteinDataTable):
        self.raw_outputs = outputs
        self.obs = obs

        self.n_samples = len(outputs)

        self.embeddings = np.array([output[0][0] for output in outputs]).reshape(self.n_samples, -1)

    def get_hidden_states(self, layer: Optional[int] = None):
        if layer is None:
            layer = -1

        return self.embeddings


class CausalLMOutput(object):
    def __init__(self, outputs):
        self.raw_outputs = outputs

        self.generated_proteins = [output['generated_text'].strip() for output in outputs]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lib310',
 'lib310.data',
 'lib310.database',
 'lib310.database.charts',
 'lib310.machinelearning',
 'lib310.tools',
 'lib310.visualization']

package_data = \
{'': ['*']}

install_requires = \
['dask-sql>=2022.11.0',
 'dask>=2022.11.0',
 'db-dtypes>=1.0.0',
 'distributed>=2022.11.0',
 'gcsfs>=2022.11.0',
 'google-cloud-bigquery>=3.2.0',
 'google-cloud>=0.34.0',
 'graphviz>=0.11',
 'matplotlib>=3.5.2',
 'numpy<1.23.0',
 'pandas>=1.4.3',
 'plotly>=5.4.0',
 'pymongo>=4.0.0',
 'rich>=12.4.4',
 'scanpy>=1.9.1',
 'seaborn>=0.11.2',
 'torch>=1.12.0']

setup_kwargs = {
    'name': 'lib310',
    'version': '0.2.41',
    'description': 'lib310 Python Package',
    'long_description': 'None',
    'author': 'Mohsen Naghipourfar',
    'author_email': 'naghipourfar@berkeley.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



from copy import deepcopy
import scanpy as sc


def cluster(model_results, method: str = 'leiden', copy: bool = False):
    adata = sc.AnnData(X=model_results.embeddings)

    if method.lower() == 'leiden':
        sc.pp.neighbors(adata)
        sc.tl.leiden(adata)

    if copy:
        model_results = deepcopy(model_results)
        model_results.obs['leiden'] = adata.obs['leiden'].values
        return model_results

    else:
        model_results.obs['leiden'] = adata.obs['leiden'].values



from google.cloud import bigquery
from google.cloud.bigquery import retry as r
from google.cloud.bigquery import enums
from google.api_core.exceptions import AlreadyExists, Conflict

from . import exceptions
from ._constants import FileFormat
from datetime import datetime
from . import _functions as fn


class Client(bigquery.Client):
    PROJECT = 'pfsdb3'
    CACHE_DATASET = 'cached'
    BUCKET_NAME = 'bigquery_1'

    def __init__(self):
        super(Client, self).__init__()

    def query(self,
              query: str,
              job_config=None,
              job_id: str = None,
              job_id_prefix: str = None,
              location: str = None,
              project: str = None,
              retry=r.DEFAULT_RETRY,
              timeout=r.DEFAULT_TIMEOUT,
              job_retry=r.DEFAULT_JOB_RETRY,
              api_method=enums.QueryApiMethod.QUERY):

        dry_config = bigquery.QueryJobConfig(dry_run=True, use_query_cache=False)
        dry_run = super(Client, self).query(query, dry_config, job_id, job_id_prefix, location, project, retry, timeout, job_retry, api_method)
        vol = int(dry_run.total_bytes_processed)
        # if vol > self.single_query_limit:
        #     raise exceptions.VolumeLimitException()
        self.__write_usage_log(vol)
        return super(Client, self).query(query, job_config, job_id, job_id_prefix, location, project, retry, timeout, job_retry, api_method)

    def build_temp_table(self):
        """
        making a table in the cached dataset
        :return: table
        """
        random_string = fn.get_random_string(8)
        ok = False
        table = None
        while not ok:
            try:
                table = self.create_table(f'{self.PROJECT}.{self.CACHE_DATASET}.{random_string}')
                ok = True
            except Conflict:
                random_string += fn.get_random_string(1)
        return table

    def query_to_cached_dataset(self, query: str, destination: bigquery.Table):

        config = bigquery.QueryJobConfig(allow_large_results=True)
        config.destination = destination
        config.write_disposition = bigquery.WriteDisposition().WRITE_TRUNCATE

        query_result = self.query(query, job_config=config, api_method=enums.QueryApiMethod.INSERT)
        return query_result.result()

    def export_to_gcs(self, table, name, destination_format=FileFormat.CSV):
        try:
            destination_uri = f"gs://{self.BUCKET_NAME}/{table.table_id}/{name}_*.{FileFormat.to_extension(destination_format)}"
        except KeyError:
            destination_uri = f"gs://{self.BUCKET_NAME}/{table.table_id}/{name}_*"
        dataset_ref = bigquery.DatasetReference(self.PROJECT, self.CACHE_DATASET)
        table_ref = dataset_ref.table(table.table_id)
        try:
            job_config = bigquery.job.ExtractJobConfig()
            job_config.destination_format = FileFormat.to_format(destination_format)
        except KeyError:
            raise TypeError('The destination format is not supported')
        extract_job = self.extract_table(
            table_ref,
            destination_uri,
            job_config=job_config
            # Location must match that of the source table.
            # location="US",
        )  # API request
        res = extract_job.result()  # Waits for job to complete.
        return res

    def __write_usage_log(self, vol):
        try:
            with open(self.__usage_log_file, 'a', encoding='utf-8') as f:
                f.write(f"{datetime.now()}: {vol} bytes\n")
                return True
        except Exception as e:
            return False

    @property
    def __usage_log_file(self):
        return '.usage_log'

    @property
    def single_query_limit(self):
        return 3 * 10 ** 12


""" lib310 python package """

import logging

from .lazy_loader import LazyLoader

# from ._settings import settings
# from . import database as db
# from . import data
# from . import machinelearning as ml
# from . import tools as tools
# from . import visualization as plot
db = LazyLoader('db', globals(), 'lib310.database')
data = LazyLoader('data', globals(), 'lib310.data')
# ml = LazyLoader('ml', globals(), 'lib310.machinelearning')
tools = LazyLoader('tools', globals(), 'lib310.tools')
plot = LazyLoader('plot', globals(), 'lib310.visualization')

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata
try:
    package_name = "lib310"
    __version__ = importlib_metadata.version(package_name)
except:
    __version__ = 'dev'

# settings.verbosity = logging.INFO

b10_logger = logging.getLogger("lib310")
b10_logger.propagate = False

__all__ = [
    'db', 'data', 'tools', 'plot',
]
__author__ = '310.ai team ("Saman Fekri <saman@310.ai>",Mohsen Naghipourfar <naghipourfar@berkeley.edu>", "Ismail Naderi <inaderi@310.ai>)'
